import Foundation

/**
Zwischen mehreren Benutzern geteilte Datei

- author: Johannes Neumeyer
- version: 1.0
*/
class SharedFile
{
    /** aktueller Dateiinhalt **/
    private var inhalt: String
    /** Singleton stellt sicher, dass nur ein Datei-Objekt existiert. **/
    static let sd = SharedFile()
    /** aktuelle Anzahl der Leser **/
    private var anzahlLeser: Int
    /** ist gerade ein Schreiber aktiv **/
    private var schreiberAktiv: Bool
    /** Das Monitorobjekt */
    private var monitor: NSCondition

    /**
    Konstruktor für Objekte der Klasse SharedFile
    nach dem Entwurfsmuster Singleton
    */
    private init()
    {
        inhalt = "Starttext"
        anzahlLeser = 0
        schreiberAktiv = false
        monitor = NSCondition()
    }

    /**
    Gibt das Datei-Objekt nach dem Singleton-Prinzip zurück.
     - returns: Dateiobjekt
    */
    static func SharedFileGeben() -> SharedFile
    {
        return sd
    }

    /**
    Ein Leser beginnt zu lesen.
    - returns: Dateiinhalt
    */
    func LesenBeginnen() -> String
    {
        monitor.lock()
        /*
         * hier ergänzen
         */
        monitor.unlock()
        return inhalt // Inhalt zum Lesen zurückgeben
    }

    /**
    Ein Leser beendet das Lesen.
    */
    func LesenBeenden()
    {
        monitor.lock()
        /*
         * hier ergänzen
         */
        monitor.unlock()
    }

    /**
    Ein Schreiber beginnt zu schreiben.
    - returns: Dateiinhalt
    */
    func SchreibenBeginnen() -> String
    {
        monitor.lock()
        /*
         * hier ergänzen
         */
        monitor.unlock()
        return inhalt // Inhalt zum Schreiben zurückgeben
    }

    /**
    Ein Schreiber speichert vorgenommene Änderungen.
     - parameters:
        - inhaltNeu: neuer Inhalt
    */
    func ÄnderungenSpeichern(inhaltNeu: String)
    {
        inhalt = inhaltNeu
    }

    /**
    Ein Schreiber beendet das Schreiben.
    */
    func SchreibenBeenden()
    {
        monitor.lock()
        /*
         * hier ergänzen
         */
        monitor.unlock()
    }
}

/**
Leser einer gemeinsamen Datei.

- author: Johannes Neumeyer
- version 1.0
 */
class Leser: Thread
{
    /** aktueller Dateiinhalt **/
    private var aktuellerInhalt: String
    /** Zähler zum Identifizieren der Leserthreads */
    private static var zähler: Int = 1
    /** Nummer des Threads */
    private var nummer: Int

    /**
    Konstruktor für Objekte der Klasse Leser
    */
    override init()
    {
        aktuellerInhalt = ""
        nummer = Leser.zähler
        Leser.zähler += 1
        super.init()
        self.start()
    }

    /**
    Der Leser möchte zu lesen beginnen.
    */
    override func main()
    {
        print("Leser \(nummer): gestartet")
        aktuellerInhalt = SharedFile.SharedFileGeben().LesenBeginnen()
        // Darstellung aktualisieren
        print("Leser \(nummer): " + aktuellerInhalt)
        //Zeitbedarf simulieren
        Thread.sleep(forTimeInterval: 0.1)
        LesenBeenden()
        print("Leser \(nummer): beendet")
    }

    /**
    Der Leser beendet das Lesen.
    */
    private func LesenBeenden()
    {
        SharedFile.SharedFileGeben().LesenBeenden()
    }
}

/**
Schreiber in eine gemeinsame Datei.

- author: Johannes Neumeyer
- version: 1.0
*/
class Schreiber: Thread
{
    /** aktueller Dateiinhalt **/
    private var aktuellerInhalt: String
    /** Zähler zum Identifizieren der Schreiberthreads */
    private static var zähler: Int = 1
    /** Nummer des Threads */
    private var nummer: Int

    /**
    Konstruktor für Objekte der Klasse Schreiber
    */
    override init()
    {
        aktuellerInhalt = ""
        nummer = Schreiber.zähler
        Schreiber.zähler += 1
        super.init()
        self.start()
    }

    /**
    Der Schreiber möchte zu schreiben beginnen.
     */
    override func main()
    {
        print("Schreiber \(nummer): gestartet")
        aktuellerInhalt = SharedFile.SharedFileGeben().SchreibenBeginnen()
        // Darstellung aktualisieren
        aktuellerInhalt = aktuellerInhalt  + "\nangegfügt von Schreiber: \(nummer)"
        SharedFile.SharedFileGeben().ÄnderungenSpeichern(inhaltNeu: aktuellerInhalt)
        print(aktuellerInhalt)
        //Zeitbedarf simulieren
        Thread.sleep(forTimeInterval: 0.2)
        Beenden()
        print("Schreiber \(nummer): beendet")
    }

    /**
    Der Schreiber beendet das Schreiben.
     */
    private func Beenden()
    {
        SharedFile.SharedFileGeben().SchreibenBeenden()
    }
}

while true
{
	if Int.random(in: 0 ..< 4) == 0
	{
		Schreiber()
    }
    else
    {
    	Leser()
    }
    Thread.sleep(forTimeInterval: Double.random(in: 0.0 ..< 1.2))
}

